from logging import Logger

from ingestion.nextgen.util.alteration_table import AlterationTableRow, ShortVariantGene
from ingestion.shared_util.open_maybe_gzipped import open_maybe_gzipped


def build_variant_key_from_vcf_line(line: str) -> str:
    split_line = line.strip().split("\t")
    chrom, pos, ref, alt = split_line[0], split_line[1], split_line[3], split_line[4]
    return f"{chrom}:{pos}:{ref}:{alt}"


def extract_filter_from_vcf_line(line: str) -> str:
    split_line = line.strip().split("\t")
    return split_line[6]


def replace_filter_in_line(line: str, new_filter: str) -> str:
    split_line = line.strip().split("\t")
    split_line[6] = new_filter
    return "\t".join(split_line) + "\n"


def is_line_in_alteration_table(
    line: str, short_variant_table_rows: list[AlterationTableRow[ShortVariantGene]]
) -> bool:
    """
    Returns True if the line in the VCF appears in
    the alteration table, False otherwise.

    Matching in the alteration table is less strict than in the
    VCF files; we only need to match chromosome and position.

    Also position may differ by +1 or -1, as deletion and insertion positions
    are represented differently in the VCF and the alteration table.
    """
    split_line = line.strip().split("\t")
    chrom, pos = split_line[0], int(split_line[1])

    for row in short_variant_table_rows:
        ref_chrom, ref_pos = row["gene"]["chr"], row["gene"]["pos"]

        if ref_chrom == chrom and (abs(ref_pos - pos) <= 1):
            return True

    return False


def pre_filter_somatic_vcf(
    somatic_vcf_file: str,
    somatic_vcf_snv_file: str,
    somatic_vcf_indel_file: str,
    short_variant_table_rows: list[AlterationTableRow[ShortVariantGene]],
    working_dir: str,
    log: Logger,
) -> str:
    """
    Removes all variants from the `somatic_vcf_file` that are not
    also in the `somatic_vcf_snv_file`, the `somatic_vcf_indel_file`,
    or the alteration table.

    Also updates the FILTER field in the `somatic_vcf_file` to match
    the FILTER field of the corresponding variant in the
    `somatic_vcf_snv_file` or `somatic_vcf_indel_file`.
    For variants in the alteration table, the original FILTER field is kept.
    """
    log.info("Pre-filtering somatic VCF file")

    valid_variants_with_filters: dict[str, str] = {}

    for file in [somatic_vcf_snv_file, somatic_vcf_indel_file]:
        with open_maybe_gzipped(file, "rt") as f:
            for line in f:
                if line.startswith("#"):
                    continue
                valid_variants_with_filters[build_variant_key_from_vcf_line(line)] = (
                    extract_filter_from_vcf_line(line)
                )

    log.info(f"Found {len(valid_variants_with_filters)} valid variants in the SNV and INDEL files")

    output_vcf_path = f"{working_dir}/filtered_somatic.vcf.gz"
    with (
        open_maybe_gzipped(somatic_vcf_file, "rt") as r,
        open_maybe_gzipped(output_vcf_path, "wt") as w,
    ):
        for line in r:
            if line.startswith("#"):
                w.write(line)
            else:
                key = build_variant_key_from_vcf_line(line)
                if key in valid_variants_with_filters:
                    w.write(replace_filter_in_line(line, valid_variants_with_filters[key]))
                elif is_line_in_alteration_table(line, short_variant_table_rows):
                    w.write(line)

    log.info(f"Successfully pre-filtered somatic VCF file to {output_vcf_path}")
    return output_vcf_path



from typing import TypedDict


class GeneWithLocation(TypedDict):
    gene: str
    chr: str
    start: int
    end: int


nextgen_specific_genes_with_location: list[GeneWithLocation] = [
    {"gene": "IGK", "chr": "chr2", "start": 88852034, "end": 90258119},
    {"gene": "NSD2", "chr": "chr4", "start": 1792518, "end": 1940193},
    {"gene": "CCND3", "chr": "chr6", "start": 41920534, "end": 42562008},
    {"gene": "MYC", "chr": "chr8", "start": 125309416, "end": 129673293},
    {"gene": "CCND1", "chr": "chr11", "start": 69090733, "end": 69656860},
    {"gene": "IGH", "chr": "chr14", "start": 105325507, "end": 109902208},
    {"gene": "MAF", "chr": "chr16", "start": 78428398, "end": 79615096},
    {"gene": "MAFB", "chr": "chr20", "start": 39039005, "end": 40688948},
    {"gene": "IGL", "chr": "chr22", "start": 22012552, "end": 22965858},
]
nextgen_specific_genes: set[str] = {gene["gene"] for gene in nextgen_specific_genes_with_location}


def maybe_get_nextgen_specific_gene(chr: str, position: int) -> str | None:
    for gene in nextgen_specific_genes_with_location:
        if gene["chr"] == chr and gene["start"] <= position <= gene["end"]:
            return gene["gene"]
    return None



import numpy as np
import re


def get_ihc_results(data, log) -> list[dict]:
    # Initialize trigger for logging, and a results list
    trigger = False
    ihc_results: list[dict] = []

    try:
        ihc_tests = [test for test in data.get("tests") if test["testMethodology"] == "IHC"]
    except TypeError:
        log.info("No Immunohistochemistry tests reported")
        return ihc_results

    hidden_ic = re.compile(r"[IC]")

    for test in ihc_tests:
        test_results = test["testResults"]["expressionAlteration"]

        biomarkername = test_results.get("biomarkerName", "")
        result = test_results.get("result", "")
        result_group = test_results.get("result_group", "")

        if result_group == "No Result":
            continue

        # Grab a field prefix list to hunt for tc / ic fields
        prefix_list = [x[:2] for x in list(test_results)]
        # If ic / tc counts are off, log it and try to load in minimal test info for future troubleshooting
        if (prefix_list.count("tc") + prefix_list.count("ic")) not in [0, 3, 7]:
            log.warning(
                f'IHC test {test_results["biomarkerName"]} has an unexpected pattern: ic/tc field count is {prefix_list.count("tc")+prefix_list.count("ic")}. Should be 0, 3, or 7(PD-L1).'
            )
            trigger = True
            ihc_results.append({"biomarkerName": biomarkername, "result": result})

        # Check for PD-L1 tests, work through odd handling patterns
        elif test_results["biomarkerName"][:5] == "PD-L1":
            # Tumor cell / immune cell logic
            if prefix_list.count("ic") == 3:
                ic_result = test_results.get("icResult", "")
                ic_stainpercent = test_results.get("icStainPercent", np.nan)

                ic_threshold = (
                    test_results["icThreshold"].split("or")[0].strip()
                    if test_results.get("icThreshold")
                    else ""
                )

                if prefix_list.count("tc") == 4:
                    tc_result = test_results.get("tcResult", "")
                    tc_intensity = test_results.get("tcIntensity", "")
                    tc_stainpercent = test_results.get("tcStainPercent", np.nan)
                    tc_threshold = (
                        test_results["tcThreshold"].split("or")[-1].strip()
                        if test_results.get("tcThreshold")
                        else ""
                    )
                    ihc_results.append(
                        {
                            "biomarkerName": biomarkername,
                            "result": result,
                            "tcResult": tc_result,
                            "tcIntensity": tc_intensity,
                            "tcStainPercent": float(tc_stainpercent),
                            "tcThreshold": tc_threshold,
                            "icResult": ic_result,
                            "icStainPercent": float(ic_stainpercent),
                            "icThreshold": ic_threshold,
                        }
                    )
                else:
                    ihc_results.append(
                        {
                            "biomarkerName": biomarkername,
                            "result": result,
                            "icResult": ic_result,
                            "icStainPercent": float(ic_stainpercent),
                            "icThreshold": ic_threshold,
                        }
                    )

            # PD-L1 (22c3) tests report a cpScore, and no stain or intensity metrics
            elif test_results["biomarkerName"][-6:] == "(22c3)":
                threshold = (
                    test_results["threshold"].split("or")[-1].strip()
                    if test_results.get("threshold")
                    else ""
                )

                tpScore = test_results.get("tpScore", np.nan)
                if np.isnan(float(tpScore)) == True:
                    cpScore = test_results.get("cpScore", np.nan)

                    ihc_results.append(
                        {
                            "biomarkerName": biomarkername,
                            "result": result,
                            "cpScore": float(cpScore),
                            "threshold": threshold,
                        }
                    )

                else:
                    ihc_results.append(
                        {
                            "biomarkerName": biomarkername,
                            "result": result,
                            "tpScore": float(tpScore),
                            "threshold": threshold,
                        }
                    )

                # Check if correct pattern is followed
                # stainPercent and intensity fields should be missing
                if test_results.get("intensity"):
                    if test_results["intensity"] != "":
                        log.warning(
                            f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "intensity": value of "{test_results["intensity"]}" was given when None was expected'
                        )
                        trigger = True
                if test_results.get("stainPercent"):
                    if np.isnan(test_results["stainPercent"]) == False:
                        log.warning(
                            f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "stainPercent": value of "{test_results["stainPercent"]}" was given when None was expected'
                        )
                        trigger = True

            # Some PD-L1 tests don't have icThreshold field filled out,
            # but report IC Threshold in the regular Threshold field
            elif isinstance((re.search(hidden_ic, test_results["threshold"])), re.Match):
                stainpercent = test_results.get("stainPercent", np.nan)

                # IC grabs threshold from the front
                threshold = (
                    test_results["threshold"].split("or")[0].strip()
                    if test_results.get("threshold")
                    else ""
                )

                ihc_results.append(
                    {
                        "biomarkerName": biomarkername,
                        "result": result,
                        "stainPercent": float(stainpercent),
                        "threshold": threshold,
                    }
                )

                # Check if correct pattern is followed
                # Intensity field should be missing
                if test_results.get("intensity"):
                    if test_results["intensity"] != "":
                        log.warning(
                            f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "intensity": value of "{test_results["intensity"]}" was given when None was expected'
                        )
                        trigger = True

            # If no PD-L1 oddities are found, then report PD-L1 as a standard test
            else:
                intensity = test_results.get("intensity", "")
                stainpercent = test_results.get("stainPercent", np.nan)
                threshold = (
                    test_results["threshold"].split("or")[-1].strip()
                    if test_results.get("threshold")
                    else ""
                )

                ihc_results.append(
                    {
                        "biomarkerName": biomarkername,
                        "result": result,
                        "intensity": intensity,
                        "stainPercent": float(stainpercent),
                        "threshold": threshold,
                    }
                )

        # A growing list of tests only report biomarker and result
        elif test_results["biomarkerName"] in [
            "Mismatch Repair Status",
            "Folfox Responder Similarity",
            "ER/PR/Her2/Neu",
        ]:
            ihc_results.append({"biomarkerName": biomarkername, "result": result})
            # Check if correct pattern is followed
            # stainPercent, intensity and threshold fields should be missing
            if test_results.get("intensity"):
                if test_results["intensity"] != "":
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "intensity": value of "{test_results["intensity"]}" was given when None was expected'
                    )
                    trigger = True
            if test_results.get("stainPercent"):
                if np.isnan(test_results["stainPercent"]) == False:
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "stainPercent": value of "{test_results["stainPercent"]}" was given when None was expected'
                    )
                    trigger = True
            if test_results.get("threshold"):
                if test_results["threshold"] != "":
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "threshold": value of "{test_results["threshold"]}" was given when None was expected'
                    )
                    trigger = True

        # Caris has a list of other tests which report result, and stainPercent
        # [MLH1, MSH2, MSH6, PMS2]
        elif test_results["biomarkerName"] in ["MLH1", "MSH2", "MSH6", "PMS2"]:
            stainpercent = test_results.get("stainPercent", np.nan)

            ihc_results.append(
                {
                    "biomarkerName": biomarkername,
                    "result": result,
                    "stainPercent": float(stainpercent),
                }
            )

            # Check if correct pattern is followed
            # Intensity and threshold fields should be missing
            if test_results.get("intensity"):
                if test_results["intensity"] != "":
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "intensity": value of "{test_results["intensity"]}" was given when None was expected'
                    )
                    trigger = True

            if test_results.get("threshold"):
                if test_results["threshold"] != "":
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "threshold": value of "{test_results["threshold"]}" was given when None was expected'
                    )
                    trigger = True

        # Standard test reporting
        else:
            intensity = test_results.get("intensity", "")
            stainpercent = test_results.get("stainPercent", np.nan)
            threshold = (
                test_results["threshold"].split("or")[-1].strip()
                if test_results.get("threshold")
                else ""
            )

            ihc_results.append(
                {
                    "biomarkerName": biomarkername,
                    "result": result,
                    "intensity": intensity,
                    "stainPercent": float(stainpercent),
                    "threshold": threshold,
                }
            )

    log.info(f"Immunohistochemistry tests detected: {len(ihc_results)}")

    # Missing field pattern checking
    for test in ihc_results:
        for k, v in test.items():
            # Log if any expected fields are left blank
            if k == "cpScore" or k == "tpScore" or k == "stainPercent":
                if np.isnan(v):
                    log.warning(
                        f'IHC test {test["biomarkerName"]} has an unexpected pattern for field: {k}'
                    )
                    trigger = True

            else:
                if v == "":
                    log.warning(
                        f'IHC test {test["biomarkerName"]} has an unexpected pattern for field: {k}'
                    )
                    trigger = True

    if trigger == False:
        log.info("All IHC tests matched the expected patterns.")

    return ihc_results



import pandas as pd
from logging import Logger

from ingestion.nextgen.util.alteration_table import AlterationTableRow, CopyNumberVariantGene
from ingestion.nextgen.util.interpretation import map_interpretation


def process_cnv(
    cnv_in_file: str,
    copy_number_variant_table_rows: list[AlterationTableRow[CopyNumberVariantGene]],
    output_dir: str,
    case_id: str,
    log: Logger,
) -> str | None:
    copy_number_path_name = f"{output_dir}/{case_id}.copynumber.csv"
    sample_id = case_id

    copy_number_variant_rows: list[str] = []

    with open(cnv_in_file, "r") as f:
        cnv_rows = f.readlines()

    for row in cnv_rows[1:]:
        working = row.strip().split("\t")

        chromosome = "chr" + working[0]
        start_position = working[1]
        end_position = working[2]
        gene = working[3]
        gene_id_only = gene.split("_")[0]
        if gene_id_only.endswith("CN"):
            gene_id_only = gene_id_only[:-2]

        copy_number = working[4]
        try:
            float(copy_number)
        except ValueError:
            log.warn("Copy number value is not a number. Omitting row.")
            continue

        status = working[5]
        if status == "normal":
            status = "neutral"

        # Hard-code
        attributes = {}

        # Scrape interpretation
        interpretation = "unknown"
        for row in copy_number_variant_table_rows:
            if (
                row["gene"]["gene"] == gene_id_only
                and row["gene"]["chr"] == chromosome
                and row["gene"]["start"] <= int(start_position)
                and row["gene"]["end"] >= int(end_position)
            ):
                interpretation = map_interpretation(row["info"], log)

        copy_number_variant_rows.append(
            f"{sample_id},{gene_id_only},{copy_number},{status},{attributes},{chromosome},{start_position},{end_position},{interpretation}\n"
        )

    if not copy_number_variant_rows:
        log.info(f"Ignoring empty copy number file {cnv_in_file}")
        return None

    log.info(f"Saving file to {copy_number_path_name}")
    with open(copy_number_path_name, "w") as f:
        f.write(
            "sample_id,gene,copy_number,status,attributes,chromosome,start_position,end_position,interpretation\n"
        )
        for cnv_text_row in copy_number_variant_rows:
            f.write(cnv_text_row)

    return copy_number_path_name



from logging import Logger

from ingestion.shared_util.gene_to_coords import gene_to_coords
from ingestion.caris.util.tests import safely_extract_tests_from_json_data


def extract_cnv(prefix, data, log: Logger):
    # Get all CNV calls into a csv
    caris_lo_keywords = {"intermediate": "gain", "amplified": "amplification", "deleted": "loss"}
    extracted_tests = safely_extract_tests_from_json_data(data)
    copy_number_variant_tests = []
    for test in extracted_tests:
        # We don't want to bring in "not detected" or wild type results
        test_name = test["testName"]
        if ("CNA" in test_name or "CND" in test_name) and "testResults" in test.keys():
            test = test["testResults"]
            if isinstance(test, dict):
                test = [test]
            for cna in test:
                if "copyNumberAlteration" in cna.keys():
                    results = cna["copyNumberAlteration"]
                    if "result" in results.keys() and results["result_group"].lower() not in [
                        "normal",
                        "no result",
                        "indeterminate",
                        "wild type",
                    ]:
                        status = results["result"].lower()
                        if status in caris_lo_keywords.keys():
                            # We only accept 2 of their results and they have to match our PHC keywords to be searchable
                            results["result"] = caris_lo_keywords[status]
                            copy_number_variant_tests.append(results)
                    elif (
                        "Exome CNA Panel - Additional Genes" in test_name
                        and "result" in results.keys()
                        and "copyNumber" in results.keys()
                    ):
                        status = results["result"].lower()

                        copy_number = 2.0
                        if results["copyNumber"]:
                            copy_number = float(results["copyNumber"])
                            cn_status = ""
                            # We only accept 2 of their results and they have to match our PHC keywords to be searchable
                            if copy_number >= 4 and copy_number < 6:
                                cn_status = "gain"
                            elif copy_number >= 6:
                                cn_status = "amplification"
                            elif copy_number < 1.3:
                                cn_status = "loss"
                            else:
                                continue
                            results["result"] = cn_status
                            copy_number_variant_tests.append(results)

    if not copy_number_variant_tests:
        return None

    # Save our results
    with open(f"{prefix}.copynumber.csv", "w") as f:
        f.write(
            "sample_id,gene,copy_number,status,attributes,chromosome,start_position,end_position,interpretation\n"
        )
        for alt in copy_number_variant_tests:
            if "genomicCoordinates" in alt.keys():
                chrom = alt["genomicCoordinates"].split(":")[1]
                coords = alt["genomicCoordinates"].split(":")[2].split("-")
            else:
                chrom, coords = gene_to_coords("GRCh37", alt["gene"])

            # Get pathogenic result
            if alt["result_group"].lower() in ["high", "mutated"]:
                interpretation = "Pathogenic"
            elif alt["result_group"].lower() in ["intermediate", "no result"]:
                interpretation = "Uncertain significance"
            else:
                interpretation = "other"

            f.write(
                ",".join(
                    [
                        prefix,
                        alt["gene"],
                        str(alt["copyNumber"]),
                        alt["result"],
                        "{}",
                        chrom,
                        coords[0],
                        coords[1],
                        f"{interpretation}\n",
                    ]
                ),
            )

    # Hard-code genome reference for Caris CNVs only
    genome_reference = "GRCh37"

    return {
        "fileName": f".lifeomic/caris/{prefix}/{prefix}.copynumber.csv",
        "sequenceType": "somatic",
        "type": "copyNumberVariant",
        "reference": genome_reference,
    }



import pandas as pd
from logging import Logger

from ingestion.caris.util.tests import safely_extract_tests_from_json_data
from ingestion.caris.util.interpretation import calculate_interpretation
from ingestion.caris.util.detect_genome_ref import detect_caris_gr


def extract_structural(prefix, data, log: Logger):
    log.info("Extracting fusion variants from json")
    extracted_tests = safely_extract_tests_from_json_data(data)
    structural_variant_tests = []
    for test in extracted_tests:
        if (
            test["platformTechnology"] in ["Transcriptome", "Hybrid Transcriptome"]
            and "testResults" in test.keys()
        ):
            for test_result in test["testResults"]:
                this_result = test_result["translocation"]

                # Result_group values:
                #     - Mutated
                #     - Normal
                #     - No Result
                #
                # We only keep Mutated. Possible results:
                #     - Fusion Detected
                #     - Likely Pathogenic Fusion
                #     - Likely Pathogenic Isoform
                #     - Pathogenic Fusion
                if this_result["result_group"].lower() not in [
                    "normal",
                    "no result",
                    "indeterminate",
                    "wild type",
                    "hlavd",
                ]:
                    structural_variant_tests.append(this_result)
    if not structural_variant_tests:
        return None

    df = pd.DataFrame(structural_variant_tests)

    def split_coords(x):
        return x.strip(":+").replace("+/", "").strip(":-").replace("-/", "")

    df["genomicBreakpoint"] = df["genomicBreakpoint"].apply(split_coords)

    df[["chromosome1", "start_position1", "chromosome2", "start_position2"]] = df[
        "genomicBreakpoint"
    ].str.split(":", expand=True)

    # Structural Variant CSV fields found in documentation here:
    # https://docs.us.lifeomic.com/user-guides/omics/data-processing/#structural-variants
    df["sample_id"] = prefix  # required str
    # df['gene1']                                 #already exists from JSON
    # df['gene2']                                 #already exists from JSON
    df["effect"] = (
        "Fusion"  # "Fusion" if "Fusion" in df['result'] else ""                   #optional str
    )
    # df['chromosome1']     = ""#firstBreak[0]#""                   #optional str
    # df['start_position1'] = ""#firstBreak[1]#""                   #optional str
    df["end_position1"] = df["start_position1"]  # optional str
    # df['chromosome2']     = ""                   #optional str
    # df['start_position2'] = ""                   #optional str
    df["end_position2"] = df["start_position2"]  # optional str
    # df['interpretation']                        #already exists from JSON
    df["sequence_type"] = df["genomicSource"]  # optional str

    # Fusions are no longer described in depth from what I can see in the new json files...
    df["in_frame"] = "Unknown"
    # To explain below: https://stackoverflow.com/a/11531402/14708230
    df.loc[df["interpretation"].str.contains("in-frame"), "in_frame"] = "Yes"

    # Utilize "result" as the interpretation, mapped to approved values
    # (if result_group is 'unclassifiedVD' we use that, because a result will not be present)
    mapped_interpretation_list = []
    for entry in zip(list(df.result_group), list(df.result)):
        if entry[0].lower() == "unclassifiedvd":
            mapped_interpretation_list.append("Uncertain significance")
        else:
            mapped_interpretation_list.append(calculate_interpretation(entry[1].lower(), log))

    df["interpretation"] = mapped_interpretation_list

    df["attributes"] = "{}"  # optional str containing JSON

    # Select columns for output
    df_out = df[
        [
            "sample_id",
            "gene1",
            "gene2",
            "effect",
            "chromosome1",
            "start_position1",
            "end_position1",
            "chromosome2",
            "start_position2",
            "end_position2",
            "interpretation",
            "sequence_type",
            "in_frame",
            "attributes",
        ]
    ]

    df_out.to_csv(f"{prefix}.structural.csv", na_rep="N/A", index=False)

    genome_reference = detect_caris_gr(structural_variant_tests, "structural", log)
    return {
        "fileName": f".lifeomic/caris/{prefix}/{prefix}.structural.csv",
        "sequenceType": "somatic",
        "type": "structuralVariant",
        "reference": genome_reference,
    }



from typing import TypedDict, Literal


class HlaResult(TypedDict):
    gene: Literal["HLA-A", "HLA-B", "HLA-C"]
    interpretation: str
    mhClass: Literal["I", "gDNA"]
    genotype: str


def extract_hla_result_from_test_result(test_result) -> HlaResult:
    gene = test_result.get("gene", "")
    if gene not in ["HLA-A", "HLA-B", "HLA-C"]:
        raise RuntimeError(f"Unknown HLA gene detected: {gene}")
    interpretation = test_result.get("result_group", "")
    mh_class = test_result.get("mhClass", "")
    if mh_class not in ["I", "gDNA"]:
        raise RuntimeError(f"Unknown MHC class detected: {mh_class}")

    genotype = test_result.get("genotype", "")

    return HlaResult(gene=gene, interpretation=interpretation, mhClass=mh_class, genotype=genotype)



from logging import Logger
import re
from typing import TypedDict, Generic, TypeVar


T = TypeVar("T")


class AlterationTableRow(Generic[T], TypedDict):
    gene: T
    type: str
    description: str
    vaf: str
    info: str


class ShortVariantGene(TypedDict):
    chr: str
    pos: int


class CopyNumberVariantGene(TypedDict):
    gene: str
    chr: str
    start: int
    end: int


class StructuralVariantGene(TypedDict):
    gene1: str
    chr1: str
    pos1: int
    gene2: str
    chr2: str
    pos2: int


base_short_variant_types: list[str] = [
    "Missense",
    "Frameshift",
    "Stop gained",
    "Stop lost",
    "Inframe deletion",
    "In-frame deletion",
    "Inframe insertion",
    "In-frame insertion",
    "Inframe",
    "In-frame",
    "Splice site",
    "Splice region",
    "Nonsense",
    "Splice acceptor",
    "Splice donor",
]


def get_short_variant_types() -> list[str]:
    # For multi-word short variant types, sometimes the spaces are not included
    short_variant_types: list[str] = []
    for short_variant_type in base_short_variant_types:
        short_variant_types.append(short_variant_type)
        if " " in short_variant_type:
            short_variant_types.append(short_variant_type.replace(" ", ""))

    return short_variant_types


def extract_all_table_lines(xml_in_file: str) -> list[str]:
    with open(xml_in_file, "r") as f:
        xml_lines = f.readlines()

    in_range_trigger = False
    table_lines: list[str] = []
    for line in xml_lines:
        if "Gene (Chr. Position, hg38)" in line:
            in_range_trigger = True
        if in_range_trigger:
            if "</Table>" in line:
                break
            table_lines.append(line)

    return table_lines


def extract_alteration_table_rows(xml_in_file: str, log: Logger) -> list[AlterationTableRow[str]]:
    table_lines = extract_all_table_lines(xml_in_file)
    # Remove completely empty lines
    table_lines = [line for line in table_lines if line.strip() != ""]

    table_row_lines: list[list[str]] = []
    current_row: list[str] = []
    for line in table_lines:
        if line.strip() == "</TR>":
            if current_row:
                table_row_lines.append(current_row)
                current_row = []
        line = re.sub(r"<\/?T.\/?>", "", line).strip()
        if line and line != "p.":
            current_row.append(line)

    alteration_table_rows: list[AlterationTableRow[str]] = []

    # Skip the first row which is the header
    for row in table_row_lines[1:]:
        # Sometimes the alteration table is "empty", in which case the `type` column will only contain "NA" values
        if row[1] == "NA":
            continue
        alteration_table_rows.append(
            {
                "gene": row[0],
                "type": row[1],
                "description": row[2],
                "vaf": row[3],
                # Sometimes the info column is empty, so we need to check if it actually exists
                # So far, it seems like rows with empty "info" columns are generally not useful for us
                # and the data in them will not be used anywhere, so we just fill in an empty string
                "info": row[4] if len(row) > 4 else "",
            }
        )

    return alteration_table_rows


def parse_short_variant_gene(gene: str) -> ShortVariantGene:
    pattern = r"^.*\((?P<chr>chr\d+|chrX|chrY):(?P<pos>\d+).*$"
    match = re.match(pattern, gene)
    if not match:
        raise RuntimeError(f"Failed to parse gene field for short variant")
    return {"chr": match.group("chr"), "pos": int(match.group("pos"))}


def parse_copy_number_variant_gene(gene: str) -> CopyNumberVariantGene:
    pattern = r"^(?P<gene>[A-Z1-9]*).*?\((?P<chr>chr\d+|chrX|chrY):(?P<start>\d+)_(?P<end>\d+)\).*$"
    match = re.match(pattern, gene)
    if not match:
        raise RuntimeError(f"Failed to parse gene field for copy number variant")
    return {
        "gene": match.group("gene"),
        "chr": match.group("chr"),
        "start": int(match.group("start")),
        "end": int(match.group("end")),
    }


def parse_structural_variant_gene(gene: str) -> StructuralVariantGene:
    pattern = r"^\*?(?P<gene1>[A-Z1-9]*)(-|\/|:+)(?P<gene2>[A-Z1-9]*).*\(.*(?P<chr1>chr\d+|chrX|chrY):(?P<pos1>\d+).*;.*(?P<chr2>chr\d+|chrX|chrY):(?P<pos2>\d+).*\).*$"
    match = re.match(pattern, gene)
    if not match:
        raise RuntimeError(f"Failed to parse gene field for structural variant")
    return {
        "gene1": match.group("gene1"),
        "chr1": match.group("chr1"),
        "pos1": int(match.group("pos1")),
        "gene2": match.group("gene2"),
        "chr2": match.group("chr2"),
        "pos2": int(match.group("pos2")),
    }


def extract_variant_table_rows_and_hyperdiploidy(xml_in_file: str, log: Logger) -> tuple[
    list[AlterationTableRow[ShortVariantGene]],
    list[AlterationTableRow[CopyNumberVariantGene]],
    list[AlterationTableRow[StructuralVariantGene]],
    list[str] | None,
]:
    alteration_table_rows = extract_alteration_table_rows(xml_in_file, log)

    short_variant_rows: list[AlterationTableRow[ShortVariantGene]] = []
    copy_number_rows: list[AlterationTableRow[CopyNumberVariantGene]] = []
    structural_variant_rows: list[AlterationTableRow[StructuralVariantGene]] = []
    hyperdiploidy_chromosomes: list[str] | None = None

    short_variant_types = get_short_variant_types()

    for row in alteration_table_rows:
        if row["type"] in short_variant_types:
            short_variant_rows.append({**row, "gene": parse_short_variant_gene(row["gene"])})
        elif row["type"] == "CNV":
            copy_number_rows.append({**row, "gene": parse_copy_number_variant_gene(row["gene"])})
        elif row["type"] == "Translocation":
            structural_variant_rows.append(
                {**row, "gene": parse_structural_variant_gene(row["gene"])}
            )
        elif row["type"] == "Hyperdiploidy":
            hyperdiploidy_chromosomes = re.findall(r"\d+", row["gene"])

    return short_variant_rows, copy_number_rows, structural_variant_rows, hyperdiploidy_chromosomes



from typing import TypedDict


class TumorMutationBurden(TypedDict):
    """A partial representation of the Tumor Mutation Burden in the Caris JSON file"""

    mutationBurdenCall: str
    mutationBurdenScore: float | str
    mutationBurdenUnit: str


class ParsedTumorMutationBurden(TypedDict):
    tmb: str
    tmbScore: float


def parse_tumor_mutation_burden(tmb_dict: TumorMutationBurden) -> ParsedTumorMutationBurden | None:
    tmb = tmb_dict["mutationBurdenCall"].lower()
    tmb_score = tmb_dict["mutationBurdenScore"]

    if not tmb_score or tmb_score == "Unknown":
        return None

    tmb_unit = tmb_dict["mutationBurdenUnit"]

    # Convert from their format, which is "21 per Mb"
    if tmb_unit == "Mutations/Megabase" and isinstance(tmb_score, str):
        if ">150" in tmb_score:
            tmb_score = 151.0
        else:
            tmb_score = float(tmb_score.split(" per")[0])
    elif isinstance(tmb_score, str):
        tmb_score = float(tmb_score)

    return {"tmb": tmb, "tmbScore": tmb_score}



from lifeomic_logging import scoped_logger
from typing import Any, TypedDict
from ruamel.yaml import YAML

from ingestion.nextgen.util.alteration_table import extract_variant_table_rows_and_hyperdiploidy
from ingestion.nextgen.util.pre_filter_somatic_vcf import pre_filter_somatic_vcf
from ingestion.nextgen.util.process_cnv import process_cnv
from ingestion.nextgen.util.process_manifest import process_manifest
from ingestion.nextgen.util.process_structural import process_structural
from ingestion.nextgen.util.process_vcf import process_vcf


class VendorFiles(TypedDict):
    somaticSvVcfFile: str
    somaticVcfFile: str
    somaticVcfSnvFile: str
    somaticVcfIndelFile: str
    germlineVcfFile: str
    pdfFile: str
    somaticBamFile: str
    somaticCnvTxtFile: str
    germlineBamFile: str
    xmlFile: str


def process(
    account_id: str,
    project_id: str,
    vendor_files: VendorFiles,
    local_output_dir: str,
    source_file_id: str,
    case_id: str,
    ingestion_id: str,
) -> dict[str, Any]:
    log_context = {
        "accountId": account_id,
        "projectId": project_id,
        "archiveFileId": source_file_id,
        "caseId": case_id,
        "ingestionId": ingestion_id,
    }
    with scoped_logger(__name__, log_context) as log:
        (
            short_variant_table_rows,
            copy_number_variant_table_rows,
            structural_variant_table_rows,
            hyperdiploidy_chromosomes,
        ) = extract_variant_table_rows_and_hyperdiploidy(vendor_files["xmlFile"], log)
        cnv_path_name = process_cnv(
            vendor_files["somaticCnvTxtFile"],
            copy_number_variant_table_rows,
            local_output_dir,
            case_id,
            log,
        )
        structural_path_name, translocations = process_structural(
            vendor_files["somaticSvVcfFile"],
            structural_variant_table_rows,
            local_output_dir,
            case_id,
            log,
        )
        manifest = process_manifest(
            vendor_files["xmlFile"],
            source_file_id,
            case_id,
            bool(cnv_path_name),
            bool(structural_path_name),
            translocations,
            hyperdiploidy_chromosomes,
            log,
        )
        pre_filtered_somatic_vcf_path = pre_filter_somatic_vcf(
            vendor_files["somaticVcfFile"],
            vendor_files["somaticVcfSnvFile"],
            vendor_files["somaticVcfIndelFile"],
            short_variant_table_rows,
            local_output_dir,
            log,
        )
        somatic_vcf_meta_data = process_vcf(
            pre_filtered_somatic_vcf_path,
            local_output_dir,
            case_id,
            "somatic",
            short_variant_table_rows,
            log=log,
        )
        germline_vcf_meta_data = process_vcf(
            vendor_files["germlineVcfFile"],
            local_output_dir,
            case_id,
            "germline",
            short_variant_table_rows,
            log,
        )

    manifest_path_name = f"{local_output_dir}/{case_id}.ga4gh.genomics.yml"
    log.info(f"Saving file to {manifest_path_name}")
    with open(manifest_path_name, "w") as file:
        yaml = YAML()
        yaml.dump(manifest, file)

    # Hard-code genome reference for nextgen
    genome_reference = "GRCh38"

    nextgen_metadata = {
        "manifest_path_name": manifest_path_name,
        "somatic_vcf_meta_data": somatic_vcf_meta_data,
        "somatic_genome_reference": genome_reference,
        "germline_vcf_meta_data": germline_vcf_meta_data,
        "germline_genome_reference": genome_reference,
    }
    if cnv_path_name:
        nextgen_metadata["cnv_path_name"] = cnv_path_name
        nextgen_metadata["cnv_genome_reference"] = genome_reference
    if structural_path_name:
        nextgen_metadata["structural_path_name"] = structural_path_name
        nextgen_metadata["structural_genome_reference"] = genome_reference

    return nextgen_metadata



import pandas as pd

from logging import Logger
from ingestion.caris.util.detect_genome_ref import detect_caris_gr


# Take the caris RNA information from the provided tsv for ingestion and input to TSO/pcann
# We are not guaranteed rnaseq results FYI
def convert_tsv_to_rgel(prefix, files, log: Logger):
    if not "tsv" in files:
        return None
    tsv_file = files["tsv"]
    log.info(f"RNA TPM file found. Converting to RGEL: {tsv_file}")

    df = pd.read_table(tsv_file, comment="#", header=None)
    df.rename(columns={0: "gene_id", 1: "expression"}, inplace=True)

    df["sample_id"] = prefix
    df["gene_name"] = df["gene_id"]
    df["raw_count"] = ""
    df["attributes"] = "{}"
    df["is_normalized"] = "True"
    df["expression_unit"] = "tpm"

    df.drop_duplicates(inplace=True)
    # Select columns for output
    df_out = df[
        [
            "sample_id",
            "gene_id",
            "gene_name",
            "expression",
            "raw_count",
            "attributes",
            "is_normalized",
            "expression_unit",
        ]
    ]

    df_out.to_csv(f"{prefix}.expression.rgel.gz", compression="gzip", na_rep="", index=False)

    headers = []
    with open(tsv_file) as f:
        for line in f.readlines()[:10]:
            headers.append(line.strip())
    genome_reference = detect_caris_gr(headers, "expression", log)

    return {
        "fileName": f".lifeomic/caris/{prefix}/{prefix}.expression.rgel.gz",
        "sequenceType": "somatic",
        "type": "expression",
        "reference": genome_reference,
    }

    return None



def safely_extract_tests_from_json_data(data: dict) -> list[dict]:
    tests = data["tests"]

    # It seems like Caris generates the JSON file from XML.
    # Because of this, sometimes a `tests` with one test gets converted to a dict.
    # We need it as a list so we can safely iterate over it.
    if isinstance(tests, dict):
        tests = [tests]

    return tests



from logging import Logger
import re


def parse_pattern(pattern: str, line: str, name: str) -> str:
    regex = re.compile(pattern)
    match = regex.match(line)

    if not match:
        raise ValueError(f"Could not parse {name} from line")

    return match.group(1).strip()


def parse_indication(line: str) -> str:
    return parse_pattern(r"^.*Reason for Referral:(.*?)(Patient|<).*$", line, "indication")


def parse_ordering_md(line: str) -> str:
    return parse_pattern(r"^.*Physician Name:(.*?)(Reason|<).*$", line, "ordering MD")


def parse_patient_name(line: str) -> str:
    return parse_pattern(r"^.*Patient Name: (.*?)(Accession|<).*$", line, "patient name")


def parse_sample_number(line: str) -> str:
    return parse_pattern(r"^.*Specimen #: (\d*-?R?) .*$", line, "sample number")


def parse_body_site(line: str) -> str:
    return parse_pattern(r"^.*Specimen:(.*?)(Age|Birthdate|<).*$", line, "body site")


def parse_report_id(line: str) -> str:
    return parse_pattern(r"^.*Accession #: (.*?) .*$", line, "report ID")


def parse_report_date_single_line(line: str) -> str:
    return parse_pattern(
        r"^.*Diagnostic Genomics Laboratory.*(\d{2}\/\d{2}\/\d{4}).*$", line, "report date"
    )


def parse_report_date_multiline(patient_info_lines: list[str]) -> str:
    in_range_trigger = False

    for line in patient_info_lines:
        if "Laboratory" in line:
            in_range_trigger = True
            continue
        if in_range_trigger:
            formatted_line = re.sub(r"<\/?T.\/?>", "", line).strip()
            if not formatted_line:
                continue
            return parse_pattern(
                r"^.*(\d{2}\/\d{2}\/\d{4}).*$", formatted_line, "report date from multiline"
            )

    raise ValueError("Could not parse report date from lines")


def parse_report_date(patient_info_lines: list[str], log: Logger) -> str:
    """
    Typically, the report date is in a form like:
    ```
    Diagnostic Genomics Laboratory 01/01/2021
    ```

    However, sometimes the date is split across multiple lines, like:
    ```
    Diagnostic Genomics Laboratory
    ...random empty lines or lines with only tags...
    01/01/2021
    ```
    This function attempts to first parse the date from a single line, and if that fails,
    it will attempt to parse it from multiple lines.
    """
    for line in patient_info_lines:
        if "Laboratory" in line:
            try:
                report_date = parse_report_date_single_line(line)
                return report_date
            except ValueError:
                log.warning("Could not parse report date from single line")
                break

    return parse_report_date_multiline(patient_info_lines)



import re
from logging import Logger

from ingestion.caris.util.hla import extract_hla_result_from_test_result
from ingestion.caris.util.ihc import get_ihc_results
from ingestion.caris.util.specimen_details import extract_and_parse_specimen_details
from ingestion.caris.util.tests import safely_extract_tests_from_json_data
from ingestion.caris.util.tmb import parse_tumor_mutation_burden

MSI_MAPPINGS = {
    "low": "low",
    "stable": "stable",
    "high": "high",
    "indeterminate": "indeterminate",
    "equivocal": "indeterminate",
}


def get_report_date(test_details, log: Logger) -> str:
    try:
        test_details["approvalInformation"]["approveDate"].split()[0]
    except KeyError:
        log.info("JSON does not contain approval information")
        return ""

    else:
        return test_details["approvalInformation"]["approveDate"].split()[0]


def get_ordering_md_name(physician_details) -> str:
    return (
        f'{physician_details["lastName"]}, {physician_details["firstName"]}'
        if physician_details.get("lastName") and physician_details.get("firstName")
        else ""
    )


def get_ordering_md_npi(physician_details) -> int:
    return physician_details.get("npi", "")


def get_med_facil_name(physician_details) -> str:
    return physician_details.get("organization", "")


def get_med_facil_id(physician_details) -> int:
    return physician_details.get("sourceID", "")


def get_report_id(test_details) -> str:
    return test_details.get("labReportID", "")


def get_physician_details(data) -> dict:
    return data.get("physicianInformation", {})


def get_test_type(data):
    if re.search(r"liquidBiopsy", data):
        return "Assure"
    if re.search(r"Hybrid_Transcriptome", data):
        return "MI Profile - Hybrid"
    return "MI Profile"


# In the array of Caris test entries, there are only some that we care about
def is_valid_test_entry(test: dict):
    if "test_cancellation_reason" in test:
        return False

    search_expressions = [r"clinical genes", r"liquid germline", r"lp exome variants"]

    is_seq = test.get("testMethodology") == "Seq"
    has_matching_name = any(
        re.search(expr, test["testName"], re.IGNORECASE) for expr in search_expressions
    )

    return is_seq and has_matching_name


# Build up the manifest iteratively because almost everything is optional
def extract_metadata(data, prefix, files, source_file_id, log: Logger) -> dict:
    is_test_cancelled_permit_vcf_skip = False
    metadata = {}

    test_details = data["testDetails"]

    specimen_details = extract_and_parse_specimen_details(data, log)

    physician_details = get_physician_details(data)

    metadata["testType"] = get_test_type(str(data))

    # Get date of collected and received for the specimen
    metadata["receivedDate"] = specimen_details["receivedDate"]
    metadata["collDate"] = specimen_details["collDate"]
    if specimen_details["specimenId"]:
        metadata["specimenIds"] = [specimen_details["specimenId"]]
    if specimen_details["specimenTypes"]:
        metadata["specimenTypes"] = specimen_details["specimenTypes"]
    metadata["reportDate"] = get_report_date(test_details, log)

    # Get the date without the time
    metadata["indexedDate"] = metadata["reportDate"]

    patient = data["patientInformation"]
    metadata["bodySiteSystem"] = "http://lifeomic.com/fhir/sequence-body-site"
    metadata["reportID"] = get_report_id(test_details)
    metadata["mrn"] = patient["mrn"]
    metadata["patientLastName"] = patient["lastName"]

    metadata["patientDOB"] = patient["dob"]

    # Get physician info - ordering name, NPI, and facility
    metadata["medFacilName"] = get_med_facil_name(physician_details)
    metadata["medFacilID"] = get_med_facil_id(physician_details)
    metadata["orderingMDName"] = get_ordering_md_name(physician_details)
    metadata["orderingMDNPI"] = get_ordering_md_npi(physician_details)

    metadata["indicationSystem"] = "http://lifeomic.com/fhir/sequence-indication"
    metadata["indication"] = patient["diagnosis"]
    metadata["indicationDisplay"] = metadata["indication"]

    metadata["bodySite"] = specimen_details["bodySite"]
    metadata["bodySiteDisplay"] = metadata["bodySite"]
    metadata["sourceFileId"] = source_file_id
    pdf = files["pdf"]
    metadata["reportFile"] = f".lifeomic/caris/{prefix}/{pdf}"

    # Some patients do not have an MRN
    patientInfo = (
        {
            "lastName": metadata["patientLastName"],
            "dob": metadata["patientDOB"],
            "firstName": patient["firstName"],
            "gender": patient["gender"].lower(),
            "identifiers": [
                {
                    "codingCode": "MR",
                    "codingSystem": "http://hl7.org/fhir/v2/0203",
                    "value": metadata["mrn"],
                }
            ],
        }
        if metadata["mrn"]
        else {
            "lastName": metadata["patientLastName"],
            "dob": metadata["patientDOB"],
            "firstName": patient["firstName"],
            "gender": patient["gender"].lower(),
        }
    )

    # Ensure no null entries
    metadata["patientInfo"] = {k: v for k, v in patientInfo.items() if v}
    metadata.update({"name": "Caris", "reference": "GRCh37"})
    metadata["hlaResults"] = []

    # Now find the test information

    tests = safely_extract_tests_from_json_data(data)

    # if not sufficient quantity we won't have test results
    if test_details["reportType"] != "QNS":
        for test in tests:
            if "test_cancellation_reason" in test:
                if test["test_cancellation_reason"] == "Quantitation quantity not sufficient":
                    # capture cancellation reason before bailing
                    # this is so we can generate an empty vcf so present biomarkers are
                    # still ingested: https://lifeomic.atlassian.net/browse/PHC-5748
                    is_test_cancelled_permit_vcf_skip = True

            if not is_valid_test_entry(test):
                continue
            # Sometimes, if there is only a single test result,
            # Caris will set it as a dict instead of a list
            test_results = (
                [test["testResults"]]
                if isinstance(test["testResults"], dict)
                else test["testResults"]
            )
            for info in test_results:
                if "tumorMutationBurden" in info.keys():
                    parsed_tmb = parse_tumor_mutation_burden(info["tumorMutationBurden"])
                    if not parsed_tmb:
                        continue
                    metadata["tmb"] = parsed_tmb["tmb"]
                    metadata["tmbScore"] = parsed_tmb["tmbScore"]
                elif "microsatelliteInstability" in info.keys():
                    # if the key isn't found we will get an error during manifest processing
                    # it would be better to fail here, i.e. fail fast, but our alerting
                    # is much better at the manifest level so doing a default value for now
                    msi_key = info["microsatelliteInstability"]["msiCall"].lower()
                    if msi_key == "not detected":
                        continue
                    if msi_key in MSI_MAPPINGS:
                        metadata["msi"] = MSI_MAPPINGS[msi_key]
                    else:
                        metadata["msi"] = msi_key
                elif "genomicLevelHeterozygosity" in info.keys():
                    loh_status = info["genomicLevelHeterozygosity"]["result"].lower()
                    loh_score = info["genomicLevelHeterozygosity"]["LOHpercentage"]
                    # This comes out as a string, convert to integer for proper ingestion
                    metadata["lossOfHeterozygosityScore"] = int(loh_score)
                    metadata["lossOfHeterozygosityStatus"] = (
                        "qns" if loh_status == "quality not sufficient" else loh_status
                    )
                elif "genomicScarScore" in info.keys():
                    hrd_status = info["genomicScarScore"]["result"].lower()
                    hrd_score = info["genomicScarScore"]["score"]
                    # This comes out as a string, convert to integer for proper ingestion
                    metadata["hrdStatus"] = (
                        "qns" if hrd_status == "quality not sufficient" else hrd_status
                    )
                elif "genomicAlteration" in info and "HLA" in info["genomicAlteration"].get(
                    "biomarkerName", ""
                ):
                    metadata["hlaResults"].append(
                        extract_hla_result_from_test_result(info["genomicAlteration"])
                    )

    # Add IHC test results to manifest
    metadata["ihcTests"] = get_ihc_results(data, log)

    # Add in the additional resources as linkable
    metadata["resources"] = []
    ingest_files = [
        "nrm.vcf.gz",
        "tmp",
        "ga4gh.genomics.yml",
        "runner",
        "yml",
        "ga4gh.yml",
        "copynumber.csv",
        "structural.csv",
        "pdf",
    ]
    for ext, filename in files.items():
        if ext not in ingest_files:
            if ext != "bam" and "fastq" not in ext:
                metadata["resources"].append({"fileName": f".lifeomic/caris/{prefix}/{filename}"})
            else:
                for f in files[ext]:
                    metadata["resources"].append({"fileName": f".lifeomic/caris/{prefix}/{f}"})
                    if ext == "bam":
                        metadata["resources"].append(
                            {"fileName": f".lifeomic/caris/{prefix}/{f}.bai"}
                        )
            # If we got RNAseq results let us also make json files available
            if ext == "tsv":
                metadata["resources"].append(
                    {"fileName": f".lifeomic/caris/{prefix}/{prefix}.expression.cancerscope.json"}
                )
                metadata["resources"].append(
                    {"fileName": f".lifeomic/caris/{prefix}/{prefix}.expression.pcann.json"}
                )

    active_metadata = {k: v for k, v in metadata.items() if v is not None}
    return (active_metadata, is_test_cancelled_permit_vcf_skip)



import os
from pathlib import Path
from ingestion.vcf_standardization.standardize import standardize_vcf
from lifeomic_logging import scoped_logger


def process(vcf_file: str, source_file_id: str, out_path: str, case_id: str) -> dict:
    with scoped_logger(__name__) as log:
        # TODO: do we need to take in + process the manifest file here?

        # Process VCF
        base_vcf_file = os.path.basename(vcf_file)
        vcf_out = base_vcf_file.replace(".vcf", ".modified.vcf")
        vcf_final = base_vcf_file.replace(".vcf", ".modified.nrm.filtered.vcf")
        if not vcf_final.endswith(".gz"):
            vcf_final = vcf_final + ".gz"

        # Assuming Nebula VCFs are germline
        sample_name = f"germline_{case_id}"
        vcf_line_count = standardize_vcf(
            vcf_file, vcf_out, out_path, sample_name, log, compression=True
        )

        # Create a basic manifest for the Nebula VCF
        manifest = {
            "testType": "Nebula",
            "sourceFileId": source_file_id,
            "reference": "GRCh38",  # Assuming GRCh38, adjust as needed
            "resources": [{"fileName": f".lifeomic/nebula/{case_id}/{base_vcf_file}"}],
            "files": [
                {
                    "fileName": f".lifeomic/nebula/{case_id}/{vcf_final}",
                    "sequenceType": "germline",
                    "type": "shortVariant",
                }
            ],
        }

        case_metadata = {
            "test_type": "Nebula",
            "vcf_line_count": vcf_line_count,
            "case_id": case_id,
            "germline_genome_reference": manifest["reference"],
        }

        return case_metadata, manifest



import gzip
from typing import cast, TextIO


def open_maybe_gzipped(filename: str, mode: str = "rt") -> TextIO:
    return cast(
        TextIO, gzip.open(filename, mode) if filename.endswith(".gz") else open(filename, mode)
    )



import re
from logging import Logger
from typing import Any

import ingestion.nextgen.util.manifest_helpers as manifest_helpers


def transform_date(date: str):
    """mm/dd/yyyy -> yyyy-mm-dd"""
    date_list = date.split("/")
    new_date = f"{date_list[2]}-{date_list[0]}-{date_list[1]}"
    return new_date


def search_and_grab(array: list, search_item: str, grab_index: int):
    """Searches through an array for a string, and returns the item at the specified index after the search item"""
    return array[array.index([i for i in array if re.search(search_item, i)][0]) + grab_index]


def extract_xml_text(xml_in_file: str) -> list[str]:
    with open(xml_in_file, "r") as f:
        xml_lines = f.readlines()

    in_range_trigger = False
    patient_info_lines = []
    for line in xml_lines:
        if "Diagnostic Genomics Laboratory " in line and "bookmark" not in line:
            in_range_trigger = True
        if "PCM Result:" in line:
            in_range_trigger = False
            break
        if in_range_trigger == True:
            patient_info_lines.append(line.strip())

    return patient_info_lines


def extract_interpretation_text(xml_in_file: str) -> list[str]:
    with open(xml_in_file, "r") as f:
        xml_lines = f.readlines()

    in_range_trigger = False
    interpretation_lines = []
    for line in xml_lines:
        if "INTERPRETATION" in line:
            in_range_trigger = True
        if "References" in line:
            in_range_trigger = False
            break
        if in_range_trigger == True:
            interpretation_lines.append(line.strip())

    return interpretation_lines


def get_cell_purity(interpretation_lines: list):
    cell_purity_value = None
    for line in interpretation_lines:
        if "CD138+" in line:
            search_result = re.search(r"\d\d\.\d\d\%", line)
            if search_result:
                cell_purity_value = float(search_result.group(0)[:-2])
                return cell_purity_value
    if not cell_purity_value:
        return float(00.00)


def extract_patient_data(patient_info_lines: list[str]):
    patient_data: dict = {}
    patient_data["patientInfo"] = {}

    for line in patient_info_lines:
        if "patientLastName" not in patient_data and "Patient Name" in line:
            patient_name = manifest_helpers.parse_patient_name(line)
            if "," not in patient_name:
                raise ValueError("Could not parse patient name from line")
            last, first = patient_name.split(",")
            patient_data["patientInfo"]["firstName"] = first.strip()
            patient_data["patientInfo"]["lastName"] = last.strip()
            patient_data["patientLastName"] = last

        if "patientDOB" not in patient_data and "Birthdate" in line:
            dobArray = line.split(" ")
            dob = search_and_grab(dobArray, "Birthdate:", 1)
            patient_data["patientDOB"] = transform_date(dob)
            patient_data["patientInfo"]["dob"] = transform_date(dob)

        if "mrn" not in patient_data and "MRN #" in line:
            mrnArray = line.split(" ")
            mrn = search_and_grab(mrnArray, "MRN", 2)
            patient_data["mrn"] = mrn
            patient_data["patientInfo"]["identifiers"] = [
                {
                    "codingCode": "MR",
                    "codingSystem": "http://hl7.org/fhir/v2/0203",
                    "value": mrn,
                }
            ]

        if "gender" not in patient_data["patientInfo"] and "Gender" in line:
            genderArray = line.split(" ")
            gender = search_and_grab(genderArray, "Gender", 1)
            if gender == "F":
                gender = "female"
            elif gender == "M":
                gender = "male"
            else:
                gender = "other"
            patient_data["patientInfo"]["gender"] = gender

    return patient_data


def extract_test_data(patient_info_lines: list[str], interpretation_lines: list[str], log: Logger):
    # Initialize manifest and hard-code some values
    manifest: dict[str, Any] = {}
    manifest["testType"] = "Plasma Cell Myeloma Panel"
    manifest["name"] = "IU Diagnostic Genomics"
    manifest["reference"] = "GRCh38"

    manifest["ihcTests"] = []
    manifest["tumorTypePredictions"] = []
    manifest["orderingMDNPI"] = ""

    manifest["bodySiteSystem"] = "http://lifeomic.com/fhir/sequence-body-site"
    manifest["indicationSystem"] = "http://lifeomic.com/fhir/sequence-indication"

    manifest["medFacilID"] = ""
    manifest["medFacilName"] = "IU Health"

    report_date = manifest_helpers.parse_report_date(patient_info_lines, log)
    manifest["reportDate"] = transform_date(report_date)
    manifest["indexedDate"] = manifest["reportDate"]

    for line in patient_info_lines:
        if "collDate" not in manifest and "Collected" in line:
            collArray = line.split(" ")
            coll_date = search_and_grab(collArray, "Collected", 1)
            manifest["collDate"] = transform_date(coll_date)

        if "receivedDate" not in manifest and "Received" in line:
            recArray = line.split(" ")
            rec_date = search_and_grab(recArray, "Received", 1)
            manifest["receivedDate"] = transform_date(rec_date)

        if "reportID" not in manifest and "Accession #" in line:
            manifest["reportID"] = manifest_helpers.parse_report_id(line)

        if "sampleNumber" not in manifest and "Specimen #" in line:
            manifest["sampleNumber"] = manifest_helpers.parse_sample_number(line)

        if "orderingMDName" not in manifest and "Physician Name:" in line:
            manifest["orderingMDName"] = manifest_helpers.parse_ordering_md(line)

        if "indication" not in manifest and "Reason for Referral" in line:
            indication = manifest_helpers.parse_indication(line)
            manifest["indication"] = indication
            manifest["indicationDisplay"] = indication

        if "bodySite" not in manifest and "Specimen:" in line:
            body_site = manifest_helpers.parse_body_site(line).strip()
            manifest["bodySite"] = body_site
            manifest["bodySiteDisplay"] = body_site

    cell_purtiy = get_cell_purity(interpretation_lines)
    if cell_purtiy != 00.00:
        manifest["cellPurity"] = cell_purtiy

    return manifest


def process_manifest(
    xml_in_file: str,
    source_file_id: str,
    case_id: str,
    include_copy_number: bool,
    include_structural: bool,
    somatic_translocations: list[str],
    hyperdiploidy_chromosomes: list[str] | None,
    log: Logger,
):
    test_text = extract_xml_text(xml_in_file)
    interpretation_text = extract_interpretation_text(xml_in_file)
    manifest = extract_test_data(test_text, interpretation_text, log)
    manifest.update(extract_patient_data(test_text))

    file_prefix = f".lifeomic/nextgen/{case_id}/{case_id}"

    if hyperdiploidy_chromosomes:
        manifest["hyperdiploidyTrisomies"] = hyperdiploidy_chromosomes
    if somatic_translocations:
        manifest["somaticTranslocations"] = somatic_translocations

    manifest["reportFile"] = f"{file_prefix}.pdf"
    manifest["sourceFileId"] = source_file_id
    manifest["resources"] = []

    manifest["files"] = [
        {
            "fileName": f"{file_prefix}.modified.somatic.nrm.filtered.vcf.gz",
            "sequenceType": "somatic",
            "type": "shortVariant",
        },
        {
            "fileName": f"{file_prefix}.modified.germline.nrm.filtered.vcf.gz",
            "sequenceType": "germline",
            "type": "shortVariant",
        },
        {
            "fileName": f"{file_prefix}.somatic.updated.bam",
            "sequenceType": "somatic",
            "type": "read",
        },
        {
            "fileName": f"{file_prefix}.germline.updated.bam",
            "sequenceType": "germline",
            "type": "read",
        },
    ]
    if include_structural:
        manifest["files"].append(
            {
                "fileName": f"{file_prefix}.structural.csv",
                "sequenceType": "somatic",
                "type": "structuralVariant",
            },
        )
    if include_copy_number:
        manifest["files"].append(
            {
                "fileName": f"{file_prefix}.copynumber.csv",
                "sequenceType": "somatic",
                "type": "copyNumberVariant",
            }
        )

    return manifest



from logging import Logger
import re
from typing import TypedDict

from ingestion.shared_util.coords_to_genes import coords_to_genes
from ingestion.nextgen.util.alteration_table import AlterationTableRow, StructuralVariantGene
from ingestion.nextgen.util.interpretation import map_interpretation
from ingestion.nextgen.util.nextgen_specific_genes import maybe_get_nextgen_specific_gene
from ingestion.shared_util.open_maybe_gzipped import open_maybe_gzipped


class StructuralVariant(TypedDict):
    sample_id: str
    gene1: str
    gene2: str
    effect: str
    position1: tuple[str, str, str]
    position2: tuple[str, str, str]
    interpretation: str
    sequence_type: str
    in_frame: str
    attributes: dict


def structural_variant_to_csv_row(structural_variant: StructuralVariant) -> str:
    csv_row = ""
    for value in structural_variant.values():
        if isinstance(value, tuple):
            csv_row += ",".join(value)
        else:
            csv_row += f"{value}"
        csv_row += ","
    return f"{csv_row[:-1]}\n"


def are_variants_duplicates(sv1: StructuralVariant, sv2: StructuralVariant) -> bool:
    return (sv1["position1"] == sv2["position1"] and sv1["position2"] == sv2["position2"]) or (
        sv1["position1"] == sv2["position2"] and sv1["position2"] == sv2["position1"]
    )


def is_del_dup_or_ins(variant: list[str]) -> bool:
    return any([x in variant[2] for x in ["MantaDEL", "MantaDUP", "MantaINS"]])


def get_center_position(start_position: str, end_position: str) -> int:
    """
    Calculate the center position of a variant based on its start and end positions, useful for finding genes.
    """
    return int((int(start_position) + int(end_position)) / 2)


def process_structural(
    structural_variant_in_file: str,
    structural_variant_table_rows: list[AlterationTableRow[StructuralVariantGene]],
    output_dir: str,
    case_id: str,
    log: Logger,
) -> tuple[str | None, list[str]]:
    structural_variant_path_name = f"{output_dir}/{case_id}.structural.csv"
    sample_id = case_id

    with open_maybe_gzipped(structural_variant_in_file, "rt") as f:
        variants = [line for line in f.readlines() if not line.startswith("#")]

    structural_variants: list[StructuralVariant] = []
    formatted_translocations: set[str] = set()

    for variant in variants:
        gene1: str | None = None
        gene2: str | None = None

        working_variant = variant.strip().split("\t")

        chromosome1 = f"chr{working_variant[0]}"
        start_position1 = working_variant[1]

        if is_del_dup_or_ins(working_variant):
            end_position1 = working_variant[7].split(";")[0].split("=")[1]
            chromosome2 = chromosome1
            start_position2 = start_position1
            end_position2 = end_position1
            effect = "duplication"
            if "MantaDEL" in working_variant[2]:
                effect = "deletion"
            elif "MantaINS" in working_variant[2]:
                effect = "insertion"

            # Get genes from coordinates using center point of start and end positions
            gene1 = None
            gene2 = "N/A"

        else:
            alt = working_variant[4].strip("][TCGA").split(":")

            end_position1 = start_position1
            chromosome2 = f"chr{alt[0]}"
            start_position2 = alt[1]
            end_position2 = alt[1]
            effect = "translocation"

            gene1 = maybe_get_nextgen_specific_gene(
                chromosome1, get_center_position(start_position1, end_position1)
            )
            gene2 = maybe_get_nextgen_specific_gene(
                chromosome2, get_center_position(start_position2, end_position2)
            )

            # Maybe add this variant to the formatted translocations list
            if (gene1 == "MYC" or gene2 == "MYC") and gene1 != gene2:
                formatted_translocations.add("t(MYC)")
            elif gene1 and gene2:
                # Remove the "chr" prefix and convert to int
                chr1, chr2 = int(chromosome1[3:]), int(chromosome2[3:])
                # Don't add translocations between the same chromosome
                if chr1 == chr2:
                    continue
                # Ensure chromosomes are in ascending order
                if chr1 > chr2:
                    chr1, chr2 = chr2, chr1
                formatted_translocations.add(f"t({chr1};{chr2})")

        # Scrape interpretation
        interpretation = "unknown"
        for row in structural_variant_table_rows:
            is_match = (
                row["gene"]["chr1"] == chromosome1
                and row["gene"]["chr2"] == chromosome2
                and row["gene"]["pos1"] == int(start_position1)
                and row["gene"]["pos2"] == int(start_position2)
            )
            if not is_match:
                continue

            interpretation = map_interpretation(row["info"], log)
            # Use the gene names from the alteration table but only if they are not already set
            gene1 = gene1 if gene1 else row["gene"]["gene1"]
            gene2 = gene2 if gene2 else row["gene"]["gene2"]

        # Hard-code
        sequence_type = "Somatic"
        in_frame = "Unknown"
        attributes: dict = {}

        # If genes have not been populated from the nextgen specific genes or alteration
        # table fall back to using the default gene finding method
        if not gene1:
            gene1 = coords_to_genes(
                "GRCh38", chromosome1, get_center_position(start_position1, end_position1), log
            )
        if not gene2:
            gene2 = coords_to_genes(
                "GRCh38", chromosome2, get_center_position(start_position2, end_position2), log
            )

        structural_variants.append(
            {
                "sample_id": sample_id,
                "gene1": gene1,
                "gene2": gene2,
                "effect": effect,
                "position1": (chromosome1, start_position1, end_position1),
                "position2": (chromosome2, start_position2, end_position2),
                "interpretation": interpretation,
                "sequence_type": sequence_type,
                "in_frame": in_frame,
                "attributes": attributes,
            }
        )

    # Dedupe structural variants based on chromosome and positions
    deduped_structural_variants: list[StructuralVariant] = []
    for sv in structural_variants:
        maybe_matching_variant = next(
            (
                variant
                for variant in deduped_structural_variants
                if are_variants_duplicates(sv, variant)
            ),
            None,
        )
        if not maybe_matching_variant:
            deduped_structural_variants.append(sv)

    if not deduped_structural_variants:
        log.info(f"Ignoring empty structural variant file {structural_variant_in_file}")
        return (None, [])

    log.info(f"Saving file to {structural_variant_path_name}")
    with open(structural_variant_path_name, "w+") as f:
        f.write(
            "sample_id,gene1,gene2,effect,chromosome1,start_position1,end_position1,chromosome2,start_position2,end_position2,interpretation,sequence_type,in-frame,attributes\n"
        )
        for sv in deduped_structural_variants:
            f.write(structural_variant_to_csv_row(sv))

    log.info(f"Found {len(formatted_translocations)} translocations for genes of interest")

    return structural_variant_path_name, list(formatted_translocations)



from logging import Logger


def map_interpretation(status: str, log: Logger):
    """
    Map interpretation for structural and copy number variants
    """
    if status == "Pathogenic":
        return "Pathogenic"
    elif "VUS" in status:
        return "Uncertain significance"
    else:
        log.error(f"Failed to resolve interpretation: {status}")
        return ""


def map_vendsig(vendsig: str) -> str:
    """
    Map vendor significance for short variants
    """
    vendsig_lower = vendsig.lower()
    if vendsig_lower in ["pathogenic"]:
        return "VENDSIG=Pathogenic"
    elif vendsig_lower in [
        "likely pathogenic",
        "likelypathogenic",
        "likey pathogenic",
        "likeypathogenic",
    ]:
        return "VENDSIG=Likely pathogenic"
    elif vendsig_lower in ["vus"]:
        return "VENDSIG=Uncertain significance"
    else:
        raise RuntimeError(f"Unable to map vendor significance: {vendsig}")



from logging import Logger
from typing import TypedDict, cast, Optional


class SpecimenDetails(TypedDict, total=False):
    """A partial representation of the specimen details in the Caris JSON file"""

    specimenReceivedDate: str
    specimenCollectionDate: str
    specimenSite: str
    specimenType: str
    specimenID: str | int


class ParsedSpecimenDetails(TypedDict):
    bodySite: str
    receivedDate: str
    collDate: str
    specimenId: Optional[str]
    specimenTypes: Optional[list[str]]


def parse_specimen_details(specimen_details: list[SpecimenDetails]) -> ParsedSpecimenDetails:
    specimen_types: list[str] = []
    for specimen in specimen_details:
        specimen_type = specimen.get("specimenType")
        if specimen_type and specimen_type not in specimen_types:
            specimen_types.append(specimen_type)
    maybe_specimen_id = specimen_details[0].get("specimenID")

    return {
        "bodySite": specimen_details[0].get("specimenSite", ""),
        "receivedDate": specimen_details[0].get("specimenReceivedDate", ""),
        "collDate": specimen_details[0].get("specimenCollectionDate", ""),
        "specimenId": str(maybe_specimen_id) if maybe_specimen_id else None,
        "specimenTypes": specimen_types if specimen_types else None,
    }


def standardize_specimen_details(
    specimen_details: SpecimenDetails | list[SpecimenDetails],
    log: Logger,
) -> list[SpecimenDetails]:
    """
    Specimen details can be a single dictionary or a list of dictionaries
    This function standardizes the input to always be a list of dictionaries

    If it is a list, we expect all the specimens to have the same site and warn otherwise.
    """
    if isinstance(specimen_details, dict):
        return [specimen_details]

    # Sometimes, we have multiple specimen details
    # In this case, we expect them to all be the same and warn otherwise
    sites = {specimen.get("specimenSite") for specimen in specimen_details}

    if len(sites) > 1:
        log.warning("Multiple specimen sites found")

    return specimen_details


def extract_and_parse_specimen_details(data: dict, log: Logger) -> ParsedSpecimenDetails:
    specimen_information = data["specimenInformation"]
    specimen_details: SpecimenDetails | list[SpecimenDetails] | None = None

    # The key for the specimen details varies based on the test type
    potential_keys = [
        # Tissue case
        "tumorSpecimenInformation",
        # Liquid case
        "liquidBiopsySpecimenInformation",
    ]
    for key in potential_keys:
        if key in specimen_information:
            specimen_details = cast(
                SpecimenDetails | list[SpecimenDetails], specimen_information[key]
            )
            break

    if not specimen_details:
        raise ValueError("No specimen details found in data")

    specimen_details = standardize_specimen_details(specimen_details, log)

    return parse_specimen_details(specimen_details)
